""" Tests for generic and EVLA CAL manifests """

# TODO when the time comes: tests of ingestion manifests for each remaining ingestion type
#  as they're implemented, in separate modules so they don't get too unwieldy

import json
import logging
import shutil
import sys
from pathlib import Path

# pylint: disable=E0401, E0402, R1721, W0621

import pytest

from ingest_envoy.ingestion_manifest import (
    IngestionManifestBuilder,
    find_manifest,
)
from ingest_envoy.manifest_components import (
    IngestionManifestKey,
    InputScienceProduct,
    InputGroup,
    ManifestParameters,
    OutputScienceProduct,
    AncillaryProduct,
    OutputGroup,
    TARFILE_EXT,
    ARTIFACT_NAME,
    WEBLOG_FILENAME,
)
from ingest_envoy.utilities import (
    ScienceProductType,
    AncillaryProductType,
    Telescope,
)

# pylint: disable=E0401, E1120

# ingest_path is NOT unused! Don't let IJ remove the import!
from .conftest import (
    ingest_path,
    populate_fake_evla_cal_ingest_path,
    WANTED_FILENAMES,
    UNWANTED,
    find_example_manifest,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

FAKE_LOCATOR = "uid://evla/calibration/doo-wah-ditty-ditty-af123"

# TODO:
@pytest.mark.skip("FIXME")
def test_filters_cal_input_files(ingest_path: Path):
    """
    We'll be getting calibration/image/eb, etc. science products from a directory under
    /lustre/aoc/cluster/pipeline/{CAPO_PROFILE}/workspaces/staging

    Make sure we take -only- the files to be ingested.

    :param ingest_path: our temporary dir
    """

    populate_fake_evla_cal_ingest_path(ingest_path)
    locator = "uid://evla/calibration/twinkle-twinkle-little-quasar"
    manifest, _ = IngestionManifestBuilder(
        telescope=Telescope.EVLA.value,
        staging_source_dir=ingest_path,
        sp_type=ScienceProductType.EVLA_CAL.value,
        locator=locator,
    ).build()

    # does manifest contain all the data it should, and none of the data it shouldn't?

    assert manifest.locator == locator
    params = manifest.parameters
    assert not params.reingest and not params.ngas_ingest and not params.calibrate

    input_group = manifest.input_group
    assert len(input_group.science_products) == 1

    output_group = manifest.output_group
    assert len(output_group.science_products) == 1
    assert len(output_group.ancillary_products) == 1
    for product in output_group.ancillary_products:
        if product.filename not in WANTED_FILENAMES:
            assert product.filename.startswith(ARTIFACT_NAME) and product.filename.endswith(
                TARFILE_EXT
            )
        assert product.filename not in UNWANTED

    sp_out = output_group.science_products[0]
    assert sp_out.type == ScienceProductType.EVLA_CAL

    assert sp_out.filename in WANTED_FILENAMES
    assert sp_out.filename not in UNWANTED

    shutil.rmtree(ingest_path)


# TODO:
@pytest.mark.skip("FIXME")
def test_writes_expected_output_files(ingest_path: Path):
    """
    Did the manifest builder produce the manifest file, the weblog, and the science product tar?

    :param ingest_path: ingestion location
    :return:
    """
    populate_fake_evla_cal_ingest_path(ingest_path)
    manifest_file, manifest = IngestionManifestBuilder(
        telescope=Telescope.EVLA.value,
        staging_source_dir=ingest_path,
        locator="uid://evla/calibration/fee-fi-fo-fum-acdf23",
        sp_type=ScienceProductType.EVLA_CAL.value,
    ).build()
    assert manifest_file
    assert manifest

    ingestion_files = [file for file in ingest_path.iterdir()]

    # at a minimum, we expect the manifest, the ingestion artifact, and the science products tar
    assert len(ingestion_files) >= 3
    mf_json = find_manifest(ingest_path)
    assert mf_json
    tars = [file for file in ingestion_files if file.name.endswith(TARFILE_EXT)]
    assert len(tars) >= 2

    shutil.rmtree(ingest_path)


# TODO:
@pytest.mark.skip("FIXME")
def test_params_json_well_formed():
    """
    Make sure our ManifestParameters makes nice JSON

    :return:
    """
    params = ManifestParameters(
        telescope=Telescope.EVLA.value,
        reingest=False,
        ngas_ingest=False,
        calibrate=False,
        staging_source_dir=Path("/home/mchammer/evla/parallel-prod"),
    )

    params_json = params.to_json()
    # if we can dump it, it's good
    json.dumps(params_json)


@pytest.mark.skip("TODO")
def test_params_properly_formatted():
    """
    TODO
    :return:
    """
    raise NotImplementedError


# TODO:
@pytest.mark.skip("FIXME")
def test_input_sp_well_formed():
    """
    Make sure our InputScienceProduct makes nice JSON

    :return:
    """
    locator = "uid://evla/calibration/vanilla_heath_bar_crunch_1a23e"
    # single science product
    sp_dict = {
        "locator": locator,
    }

    sp_in = InputScienceProduct(locator=locator)
    assert sp_in.to_json() == sp_dict


# TODO:
@pytest.mark.skip("FIXME or get rid of me")
def test_input_group_well_formed():
    """
    Make sure our InputGroup makes nice JSON

    :return:
    """
    sp1 = InputScienceProduct(locator="uid://evla/execblock/coffee_heath_bar_crunch_7a23f")
    sp1_json = sp1.to_json()

    sp2 = InputScienceProduct(locator="uid://evla/execblock/mint_oreo_omg_omg")
    sp2_json = sp2.to_json()

    expected = {
        IngestionManifestKey.INPUT_GROUP.value: {
            IngestionManifestKey.SCIENCE_PRODUCTS.value: [sp1_json, sp2_json]
        }
    }
    ingroup = InputGroup(science_products=[sp1, sp2])
    actual = ingroup.to_json()

    actual = actual[IngestionManifestKey.INPUT_GROUP.value]
    expected = expected[IngestionManifestKey.INPUT_GROUP.value]
    assert actual.keys() == expected.keys()

    actual = actual[IngestionManifestKey.SCIENCE_PRODUCTS.value]
    expected = expected[IngestionManifestKey.SCIENCE_PRODUCTS.value]
    assert len(actual) == len(expected) == 2
    marvin = actual[0]
    trillian = expected[0]
    for key, _ in marvin.items():
        assert trillian[key] == marvin[key]

    marvin = actual[1]
    trillian = expected[1]
    for key, _ in marvin.items():
        assert trillian[key] == marvin[key]


# TODO:
@pytest.mark.skip("FIXME")
def test_ancillary_product_well_formed():
    """
    The JSON shouldn't contain empty fields

    :return:
    """
    ap1 = AncillaryProduct(type=AncillaryProductType.LOG_TYPE, filename="without_feathers.tar")
    expected = {"type": AncillaryProductType.LOG_TYPE.value, "filename": "without_feathers.tar"}
    actual = ap1.to_json()

    assert actual == expected


# TODO:
@pytest.mark.skip("FIXME or get rid of me")
def test_output_group_well_formed():
    """
    Make sure our OutputScienceProduct makes nice JSON

    :return:
    """
    sp1 = OutputScienceProduct(type=ScienceProductType.IMAGE, filename="see_no_evil.fits")
    sp2 = OutputScienceProduct(type=ScienceProductType.IMAGE, filename="hear_no_evil.fits")
    ap1 = AncillaryProduct(
        type=AncillaryProductType.PIPELINE_ARTIFACTS, filename="without_feathers.tar"
    )
    ap2 = AncillaryProduct(
        type=AncillaryProductType.PIPELINE_ARTIFACTS, filename="with_feathers.tar"
    )
    opg = OutputGroup(science_products=[sp1, sp2], ancillary_products=[ap1, ap2])
    opg_json = opg.to_json()
    dumped = json.dumps(opg_json)

    assert (
        dumped == '{"output_group": '
        '{"science_products": '
        '[{"type": "image", "filename": "see_no_evil.fits"}, '
        '{"type": "image", "filename": "hear_no_evil.fits"}], '
        '"ancillary_products": [{"type": "pipeline_artifacts", '
        '"filename": "without_feathers.tar"}, '
        '{"type": "pipeline_artifacts", "filename": "with_feathers.tar"}]}}'
    )


@pytest.mark.skip("TODO, or maybe not")
def test_input_group_properly_formatted():
    """
    Does the InputGroup get written to the file in the expected format?

    :return:
    """
    sp1 = InputScienceProduct(locator="uid://evla/execblock/coffee_heath_bar_crunch_7a23f")

    ingroup = InputGroup(science_products=[sp1])
    ig_dict = json.loads(ingroup.to_json())
    ig_text = json.dumps(ig_dict, indent=4)
    expected = """
  "input_group": {
    "science_products": [
      {
        "locator": "uid://evla/execblock/coffee_heath_bar_crunch_7a23f"
      }
    ]
  },
    """
    assert ig_text == expected


@pytest.mark.skip("TODO")
def test_output_group_properly_formatted():
    """
    TODO
    Does the OutputGroup get written to the file in the expected format?

    :return:
    """
    raise NotImplementedError


def test_evla_cal_manifest_matches_example(ingest_path: Path):
    """
    Given the correct parameters, manifest that matches _16B_069_cal_manifest.json
    should be generated

    :return:
    """

    expected_dir_name = "/lustre/aoc/cluster/pipeline/dsoc-dev/workspaces/staging/cal_test6"
    example = find_example_manifest("_16B_069_cal_manifest")
    with open(example, "r") as infile:
        expected_json = dict(json.load(infile).items())

    # populate ingestion path with fake files for manifest builder to find
    for filename in [
        "16B-069_sb32814386_1_001.57685.66193635417.testdate.caltables.tar",
        WEBLOG_FILENAME,
    ]:
        file = ingest_path / filename
        file.touch()

    builder = IngestionManifestBuilder(
        staging_source_dir=ingest_path,
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.EVLA_CAL.value,
        locator="uid://evla/execblock/48ba4c9d-d7c7-4a8f-9803-1115cd52459b",
    )
    _, manifest_file = builder.build()

    with open(manifest_file, "r") as infile:
        actual_json = dict(json.load(infile).items())

    actual_json[IngestionManifestKey.PARAMETERS.value][
        IngestionManifestKey.INGESTION_PATH.value
    ] = expected_dir_name

    assert actual_json == expected_json

    shutil.rmtree(ingest_path)


def populate_ingest_path_for_manifest_evla_cal_example(ingest_path: Path):
    """
    Create fake input files to match EVLA CAL manifest example

    :param ingest_path:
    :return:
    """
    weblog_file = ingest_path / "weblog.tgz"
    weblog_file.touch()
    cal_file = ingest_path / "XYZ-abc+TMN.O00.tar"
    cal_file.touch()
